#!/usr/bin/env ruby

# TCPSocketPipe.rb -- Creates I/O pipes for TCP socket tunneling.
# Copyright (C) 1999-2001 NAKAMURA, Hiroshi

# This application is copyrighted free software by NAKAMURA, Hiroshi.
# You can redistribute it and/or modify it under the same term as Ruby.

RCS_ID = %q$Id: TCPSocketPipe.rb,v 1.12 2001/07/13 04:16:55 nakahiro Exp $

# Ruby bundled library
require 'socket'
require 'getopts'

# Extra library
# http://raa.ruby-lang.org/list.rhtml?name=devel-logger
require 'devel/logger'

module Dump
  # Written by Arai-san and published at [ruby-list:31987].
  # http://blade.nagaokaut.ac.jp/cgi-bin/scat.rb/ruby/ruby-list/31987
  def hexdump(str)
    offset = 0
    result = []
    while raw = str.slice(offset, 16) and raw.length > 0
      # data field
      data = ''
      for v in raw.unpack('N* a*')
	if v.kind_of? Integer
	  data << sprintf("%08x ", v)
	else
	  v.each_byte {|c| data << sprintf("%02x", c) }
	end
      end
      # text field
      text = raw.tr("\000-\037\177-\377", ".")
      result << sprintf("%08x  %-36s  %s", offset, data, text)
      offset += 16
      # omit duplicate line
      if /^(#{ Regexp.quote(raw) })+/n =~ str[offset .. -1]
	result << sprintf("%08x  ...", offset)
	offset += $&.length
	# should print at the end
	if offset == str.length
	  result << sprintf("%08x  %-36s  %s", offset-16, data, text)
	end
      end
    end
    result
  end
  module_function :hexdump
end


class TCPSocketPipe < Devel::Application
  include Devel::Logger::Severity
  include Socket::Constants

  attr_accessor :dumpRequest
  attr_accessor :dumpResponse

private

  Timeout = 100			# [sec]
  ReadBlockSize = 10 * 1024	# [byte]

  class SessionPool
  public

    def each
      @pool.each do |i|
	yield i
      end
    end

    def add(serverSock, clientSock)
      @pool.push(Session.new(serverSock, clientSock))
    end

    def del(session)
      @pool.delete_if do |i|
        session.equal?(i)
      end
    end

  private

    class Session
      attr(:server)
      attr(:client)

      private

      def initialize(server = nil, client = nil)
      	@server = server
      	@client = client
      end
    end

    def initialize
      @pool = []
    end
  end

  AppName = 'TCPSocketPipe'
  ShiftAge = 0
  ShiftSize = 0

  def initialize(srcPort, destName, destPort)
    super(AppName)
    setLog(AppName + '.log', ShiftAge, ShiftSize)
    @srcPort = srcPort.to_i
    @destName = destName
    @destPort = destPort.to_i
    @dumpRequest = true
    @dumpResponse = false
    @sessionPool = SessionPool.new
  end

  def run
    @waitSock = TCPServer.new(@srcPort)
    begin
      dumpStart

      while true
        readWait = []
        @sessionPool.each do |session|
	  readWait.push(session.server).push(session.client)
        end
        readWait.unshift(@waitSock)
        readReady, writeReady, except = IO.select(readWait, nil, nil, Timeout)
        next unless readReady
        readReady.each do |sock|
	  if (@waitSock.equal?(sock))
	    newSock = @waitSock.accept
	    dumpAccept(newSock.peeraddr[2])
	    if !addSession(newSock)
      	      log(SEV_WARN, 'Closing server socket...')
	      newSock.close
	    end
	  else
	    @sessionPool.each do |session|
	      transfer(session, true) if (sock.equal?(session.server))
	      transfer(session, false) if (sock.equal?(session.client))
	    end
	  end
	end
      end
    ensure
      @waitSock.close
      dumpEnd
    end
  end

  def transfer(session, bServer)
    readSock = nil
    writeSock = nil
    if (bServer)
      readSock = session.server
      writeSock = session.client
    else
      readSock = session.client
      writeSock = session.server
    end

    readBuf = ''
    begin
      readBuf << readSock.sysread(ReadBlockSize)
    rescue EOFError
      closeSession(session)
      return
    rescue Errno::ECONNRESET
      log(SEV_INFO, "#{$!} while reading.")
      closeSession(session)
      return
    rescue
      log(SEV_WARN, "Detected an exception. Stopping ...")
      log(SEV_WARN, $@)
      closeSession(session)
      return
    end

    if (bServer)
      dumpTransferData(true, readBuf) if @dumpRequest
    else
      dumpTransferData(false, readBuf) if @dumpResponse
    end

    writeSize = 0
    while (writeSize < readBuf.size)
      begin
      	writeSize += writeSock.syswrite(readBuf[writeSize..-1])
      rescue Errno::ECONNRESET
      	log(SEV_INFO, "#{$!} while writing.")
	log(SEV_INFO, $@)
      	closeSession(session)
      	return
      rescue
	log(SEV_WARN, "Detected an exception. Stopping ...")
	log(SEV_WARN, $@)
	closeSession(session)
	return
      end
    end
  end

  def addSession(serverSock)
    begin
      clientSock = TCPSocket.new(@destName, @destPort)
    rescue
      log(SEV_ERROR, 'Create client socket failed.')
      return
    end
    @sessionPool.add(serverSock, clientSock)
    dumpAddSession
  end

  def closeSession(session)
    session.server.close
    session.client.close
    @sessionPool.del(session)
    dumpCloseSession
  end

  def dumpStart
    log(SEV_INFO, 'Started ... SrcPort=%s, DestName=%s, DestPort=%s' % [@srcPort, @destName, @destPort])
  end

  def dumpAccept(from)
    log(SEV_INFO, 'Accepted ... from ' << from)
  end

  def dumpAddSession
    log(SEV_INFO, 'Connection established.')
  end

  def dumpTransferData(isFromSrcToDestP, data)
    if isFromSrcToDestP
      log(SEV_INFO, 'Transfer data ... [src] -> [dest]')
    else
      log(SEV_INFO, 'Transfer data ... [src] <- [dest]')
    end
    dumpData(data)
  end

  def dumpData(data)
    log(SEV_INFO, "Transferred data;\n" << Dump.hexdump(data).join("\n"))
  end

  def dumpCloseSession
    log(SEV_INFO, 'Connection closed.')
  end

  def dumpEnd
    log(SEV_INFO, 'Stopped ... SrcPort=%s, DestName=%s, DestPort=%s' % [@srcPort, @destName, @destPort])
  end
end

def main
  getopts('des', 'w:', 'x:')
  srcPort = ARGV.shift
  destName = ARGV.shift
  destPort = ARGV.shift
  usage if (!srcPort or !destName or !destPort)

  # To run as a daemon...
  if $OPT_s
    exit! if fork
    Process.setsid
    exit! if fork
    STDIN.close
    STDOUT.close
    STDERR.close
  end

  app = TCPSocketPipe.new(srcPort, destName, destPort)
  app.dumpResponse = true if $OPT_d
  app.start
end

def usage
  STDERR.print <<EOM
Usage: #{$0} srcPort destName destPort

    Creates I/O pipes for TCP socket tunneling.

    srcPort .... source port# of localhost.
    destName ... hostname of a destination(name or ip-addr).
    destPort ... destination port# of the destName.

  Dump options:
    -d ......... dumps data from destination port(not dumped by default).
    -s ......... run as a daemon.

#{RCS_ID}
EOM
  exit 1
end

main if ($0 == __FILE__)
